/**
 * @file
 * VuoSceneRenderer implementation.
 *
 * @copyright Copyright © 2012–2013 Kosada Incorporated.
 * This code may be modified and distributed under the terms of the MIT License.
 * For more information, see http://vuo.org/license.
 */

#include "VuoSceneRenderer.h"
#include "VuoSceneObject.h"

#include <OpenGL/gl.h>
//#import <OpenGL/gl3.h>
/// @todo After we drop 10.6 support, switch back to gl3 and remove the below 4 lines.  See also r15430 for shader changes.
#include <OpenGL/glext.h>
/// @{
#define glGenVertexArrays glGenVertexArraysAPPLE
#define glBindVertexArray glBindVertexArrayAPPLE
#define glDeleteVertexArrays glDeleteVertexArraysAPPLE
/// @}

#include <dispatch/dispatch.h>

extern "C"
{
/// @todo uncomment when https://b33p.net/kosada/node/4018 is fixed, and add appropriate dependencies.
//const char *moduleDependencies[] = {  };
}

/**
 * GL Objects corresponding with a VuoVertices instance.
 */
struct VuoSceneRendererInternal_vertices
{
	GLuint vertexArray;
	GLuint positionBuffer;
	GLuint normalBuffer;
	GLuint tangentBuffer;
	GLuint bitangentBuffer;
	GLuint textureCoordinateBuffer;
	GLuint elementBuffer;
};

/**
 * Internal state data for a VuoSceneRenderer instance.
 */
struct VuoSceneRendererInternal
{
	dispatch_semaphore_t scenegraphSemaphore; ///< Serializes access to @c rootSceneObject.
	bool scenegraphValid;
	bool scenegraphChanged;
	VuoSceneObject rootSceneObject;

	float projectionMatrix[16]; ///< Column-major 4x4 matrix

	// Array of count VuoListGetCount_VuoVertices(rootSceneObject.vertices)
	VuoSceneRendererInternal_vertices *vertices;
};

/**
 * Creates an object for rendering a scenegraph.
 *
 * May be called from any thread (doesn't require a GL Context).
 */
VuoSceneRenderer VuoSceneRenderer_create(void)
{
	struct VuoSceneRendererInternal *sceneRenderer = (struct VuoSceneRendererInternal *)malloc(sizeof(struct VuoSceneRendererInternal));

	sceneRenderer->scenegraphSemaphore = dispatch_semaphore_create(1);
	sceneRenderer->scenegraphValid = false;
	sceneRenderer->scenegraphChanged = false;

	sceneRenderer->vertices = NULL;

	return (VuoSceneRenderer)sceneRenderer;
}

/**
 * Sets up OpenGL state on the current GL Context.
 *
 * Must be called from a thread with an active GL Context.
 */
void VuoSceneRenderer_prepareContext(VuoSceneRenderer sceneRenderer)
{
	glEnable(GL_MULTISAMPLE);

	glEnable(GL_CULL_FACE);

	glEnable(GL_DEPTH_TEST);

	glEnable(GL_BLEND);

	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

/**
 * Recalculates the projection matrix based on the specified viewport @c width and @c height.
 *
 * May be called from any thread (doesn't require a GL Context).
 */
void VuoSceneRenderer_regenerateProjectionMatrix(VuoSceneRenderer sr, unsigned int width, unsigned int height)
{
	struct VuoSceneRendererInternal *sceneRenderer = (struct VuoSceneRendererInternal *)sr;

	float fieldOfView = 45. * M_PI/180.;
	float aspectRatio = (float)width/(float)height;
	/// @todo Where to place clip planes?
	float nearClipDistance = .1;
	float farClipDistance = 10.;
	float cameraZ = 1.;

	sceneRenderer->projectionMatrix[ 0] = 1./tanf(fieldOfView);
	sceneRenderer->projectionMatrix[ 1] = 0;
	sceneRenderer->projectionMatrix[ 2] = 0;
	sceneRenderer->projectionMatrix[ 3] = 0;

	sceneRenderer->projectionMatrix[ 4] = 0;
	sceneRenderer->projectionMatrix[ 5] = aspectRatio/tanf(fieldOfView);
	sceneRenderer->projectionMatrix[ 6] = 0;
	sceneRenderer->projectionMatrix[ 7] = 0;

	sceneRenderer->projectionMatrix[ 8] = 0;
	sceneRenderer->projectionMatrix[ 9] = 0;
	sceneRenderer->projectionMatrix[10] = (farClipDistance+nearClipDistance)/(farClipDistance-nearClipDistance);
	sceneRenderer->projectionMatrix[11] = 1.;

	sceneRenderer->projectionMatrix[12] = 0;
	sceneRenderer->projectionMatrix[13] = 0;
	sceneRenderer->projectionMatrix[14] = cameraZ * sceneRenderer->projectionMatrix[10] - 2. * farClipDistance*nearClipDistance/(farClipDistance-nearClipDistance);
	sceneRenderer->projectionMatrix[15] = cameraZ;
}

/**
 * Renders the scene.
 *
 * Must be called from a thread with an active GL Context.
 */
void VuoSceneRenderer_draw(VuoSceneRenderer sr)
{
	struct VuoSceneRendererInternal *sceneRenderer = (struct VuoSceneRendererInternal *)sr;

	if (!sceneRenderer->scenegraphValid)
		return;

	dispatch_semaphore_wait(sceneRenderer->scenegraphSemaphore, DISPATCH_TIME_FOREVER);

	if (sceneRenderer->scenegraphChanged)
	{
		//fprintf(stderr, "VuoSceneRenderer_draw() scenegraphChanged\n");
		sceneRenderer->scenegraphChanged = false;

		// Upload the scenegraph.
		/// @todo Currently this only uploads the root object; need to traverse the graph.

		unsigned int verticesListCount = VuoListGetCount_VuoVertices(sceneRenderer->rootSceneObject.verticesList);
		if (sceneRenderer->vertices)
		{
			for (unsigned int i = 0; i < verticesListCount; ++i)
			{
				glDeleteVertexArrays(1, &sceneRenderer->vertices[i].vertexArray);
				glDeleteBuffers(1, &sceneRenderer->vertices[i].positionBuffer);
				glDeleteBuffers(1, &sceneRenderer->vertices[i].normalBuffer);
				glDeleteBuffers(1, &sceneRenderer->vertices[i].tangentBuffer);
				glDeleteBuffers(1, &sceneRenderer->vertices[i].bitangentBuffer);
				glDeleteBuffers(1, &sceneRenderer->vertices[i].textureCoordinateBuffer);
				glDeleteBuffers(1, &sceneRenderer->vertices[i].elementBuffer);
			}
			free(sceneRenderer->vertices);
		}
		sceneRenderer->vertices = (VuoSceneRendererInternal_vertices *)malloc(sizeof(VuoSceneRendererInternal_vertices)*verticesListCount);

		for (unsigned int i = 0; i < verticesListCount; ++i)
		{
			VuoVertices vertices = VuoListGetValueAtIndex_VuoVertices(sceneRenderer->rootSceneObject.verticesList, i);

			glGenVertexArrays(1, &sceneRenderer->vertices[i].vertexArray);
			glBindVertexArray(sceneRenderer->vertices[i].vertexArray);

			glGenBuffers(1, &sceneRenderer->vertices[i].positionBuffer);
			glBindBuffer(GL_ARRAY_BUFFER, sceneRenderer->vertices[i].positionBuffer);
			glBufferData(GL_ARRAY_BUFFER, sizeof(VuoPoint4d)*vertices.vertexCount, vertices.positions, GL_STATIC_DRAW);

			glGenBuffers(1, &sceneRenderer->vertices[i].normalBuffer);
			glBindBuffer(GL_ARRAY_BUFFER, sceneRenderer->vertices[i].normalBuffer);
			glBufferData(GL_ARRAY_BUFFER, sizeof(VuoPoint4d)*vertices.vertexCount, vertices.normals, GL_STATIC_DRAW);

			glGenBuffers(1, &sceneRenderer->vertices[i].tangentBuffer);
			glBindBuffer(GL_ARRAY_BUFFER, sceneRenderer->vertices[i].tangentBuffer);
			glBufferData(GL_ARRAY_BUFFER, sizeof(VuoPoint4d)*vertices.vertexCount, vertices.tangents, GL_STATIC_DRAW);

			glGenBuffers(1, &sceneRenderer->vertices[i].bitangentBuffer);
			glBindBuffer(GL_ARRAY_BUFFER, sceneRenderer->vertices[i].bitangentBuffer);
			glBufferData(GL_ARRAY_BUFFER, sizeof(VuoPoint4d)*vertices.vertexCount, vertices.bitangents, GL_STATIC_DRAW);

			glGenBuffers(1, &sceneRenderer->vertices[i].textureCoordinateBuffer);
			glBindBuffer(GL_ARRAY_BUFFER, sceneRenderer->vertices[i].textureCoordinateBuffer);
			glBufferData(GL_ARRAY_BUFFER, sizeof(VuoPoint4d)*vertices.vertexCount, vertices.textureCoordinates, GL_STATIC_DRAW);

			glGenBuffers(1, &sceneRenderer->vertices[i].elementBuffer);
			glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, sceneRenderer->vertices[i].elementBuffer);
			glBufferData(GL_ELEMENT_ARRAY_BUFFER, sizeof(unsigned int)*vertices.elementCount, vertices.elements, GL_STATIC_DRAW);
		}
	}


	//fprintf(stderr, "VuoSceneRenderer_draw() draw\n");
	// Draw the scenegraph.
	/// @todo Currently this only draws the root object; need to traverse the graph.
	glUseProgram(sceneRenderer->rootSceneObject.shader->glProgramName);
	{
		GLint projectionMatrixUniform = glGetUniformLocation(sceneRenderer->rootSceneObject.shader->glProgramName, "projectionMatrix");
		glUniformMatrix4fv(projectionMatrixUniform, 1, GL_FALSE, sceneRenderer->projectionMatrix);

		GLint modelviewMatrixUniform = glGetUniformLocation(sceneRenderer->rootSceneObject.shader->glProgramName, "modelviewMatrix");
		float modelviewMatrix[16];
		VuoTransform_getMatrix(sceneRenderer->rootSceneObject.transform, modelviewMatrix);
		glUniformMatrix4fv(modelviewMatrixUniform, 1, GL_FALSE, modelviewMatrix);

		VuoShader_activateTextures(sceneRenderer->rootSceneObject.shader);

		unsigned int verticesListCount = VuoListGetCount_VuoVertices(sceneRenderer->rootSceneObject.verticesList);
		for (unsigned int i = 0; i < verticesListCount; ++i)
		{
			GLint positionAttribute = glGetAttribLocation(sceneRenderer->rootSceneObject.shader->glProgramName, "position");
			glBindBuffer(GL_ARRAY_BUFFER, sceneRenderer->vertices[i].positionBuffer);
			glVertexAttribPointer(positionAttribute, 4 /* XYZW */, GL_FLOAT, GL_FALSE, sizeof(VuoPoint4d), (void*)0);
			glEnableVertexAttribArray(positionAttribute);

			GLint normalAttribute = glGetAttribLocation(sceneRenderer->rootSceneObject.shader->glProgramName, "normal");
			if (normalAttribute >= 0)
			{
				glBindBuffer(GL_ARRAY_BUFFER, sceneRenderer->vertices[i].normalBuffer);
				glVertexAttribPointer(normalAttribute, 4 /* XYZW */, GL_FLOAT, GL_FALSE, sizeof(VuoPoint4d), (void*)0);
				glEnableVertexAttribArray(normalAttribute);
			}

			GLint tangentAttribute = glGetAttribLocation(sceneRenderer->rootSceneObject.shader->glProgramName, "tangent");
			if (tangentAttribute >= 0)
			{
				glBindBuffer(GL_ARRAY_BUFFER, sceneRenderer->vertices[i].tangentBuffer);
				glVertexAttribPointer(tangentAttribute, 4 /* XYZW */, GL_FLOAT, GL_FALSE, sizeof(VuoPoint4d), (void*)0);
				glEnableVertexAttribArray(tangentAttribute);
			}

			GLint bitangentAttribute = glGetAttribLocation(sceneRenderer->rootSceneObject.shader->glProgramName, "bitangent");
			if (bitangentAttribute >= 0)
			{
				glBindBuffer(GL_ARRAY_BUFFER, sceneRenderer->vertices[i].bitangentBuffer);
				glVertexAttribPointer(bitangentAttribute, 4 /* XYZW */, GL_FLOAT, GL_FALSE, sizeof(VuoPoint4d), (void*)0);
				glEnableVertexAttribArray(bitangentAttribute);
			}

			GLint textureCoordinateAttribute = glGetAttribLocation(sceneRenderer->rootSceneObject.shader->glProgramName, "textureCoordinate");
			if (textureCoordinateAttribute >= 0)
			{
				glBindBuffer(GL_ARRAY_BUFFER, sceneRenderer->vertices[i].textureCoordinateBuffer);
				glVertexAttribPointer(textureCoordinateAttribute, 4 /* XYZW */, GL_FLOAT, GL_FALSE, sizeof(VuoPoint4d), (void*)0);
				glEnableVertexAttribArray(textureCoordinateAttribute);
			}

			{
				glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, sceneRenderer->vertices[i].elementBuffer);
				VuoVertices vertices = VuoListGetValueAtIndex_VuoVertices(sceneRenderer->rootSceneObject.verticesList, i);
				GLenum mode;
				if (vertices.elementAssemblyMethod == VuoVertices_IndividualTriangles)
					mode = GL_TRIANGLES;
				else if (vertices.elementAssemblyMethod == VuoVertices_TriangleStrip)
					mode = GL_TRIANGLE_STRIP;
				else if (vertices.elementAssemblyMethod == VuoVertices_TriangleFan)
					mode = GL_TRIANGLE_FAN;
				glDrawElements(mode, vertices.elementCount, GL_UNSIGNED_INT, (void*)0);
			}

			if (textureCoordinateAttribute >= 0)
				glDisableVertexAttribArray(textureCoordinateAttribute);
			if (bitangentAttribute >= 0)
				glDisableVertexAttribArray(bitangentAttribute);
			if (tangentAttribute >= 0)
				glDisableVertexAttribArray(tangentAttribute);
			if (normalAttribute >= 0)
				glDisableVertexAttribArray(normalAttribute);
			glDisableVertexAttribArray(positionAttribute);
		}

		VuoShader_deactivateTextures(sceneRenderer->rootSceneObject.shader);
	}
	glUseProgram(0);

	dispatch_semaphore_signal(sceneRenderer->scenegraphSemaphore);
}

/**
 * Draws all vertex normals in sceneRenderer-rootSceneObject.
 *
 * Must be called from a thread with an active GL Context.
 */
void VuoSceneRenderer_drawElement(VuoSceneRenderer sr, int element, double length)	// TODO Enum type for element to debug
{
	struct VuoSceneRendererInternal *sceneRenderer = (struct VuoSceneRendererInternal *)sr;

	if (!sceneRenderer->scenegraphValid)
		return;

	dispatch_semaphore_wait(sceneRenderer->scenegraphSemaphore, DISPATCH_TIME_FOREVER);

	unsigned int verticesListCount = VuoListGetCount_VuoVertices(sceneRenderer->rootSceneObject.verticesList);

	glLoadIdentity();

	glMatrixMode(GL_PROJECTION);
	glLoadMatrixf(sceneRenderer->projectionMatrix);


	glMatrixMode(GL_MODELVIEW);

	float modelViewMatrix[16];
	VuoTransform_getMatrix(sceneRenderer->rootSceneObject.transform, modelViewMatrix);
	glLoadMatrixf(modelViewMatrix);

	glBegin(GL_LINES);
	for (unsigned int i = 0; i < verticesListCount; i++)
	{
		VuoVertices vertices = VuoListGetValueAtIndex_VuoVertices(sceneRenderer->rootSceneObject.verticesList, i);

		for(unsigned int m = 0; m < vertices.vertexCount; m++)
		{
			VuoPoint4d v = vertices.positions[m];
			VuoPoint4d n = (VuoPoint4d){0,0,1,1};

			switch(element)
			{
				case 0:	// normals
					n = vertices.normals[m];
					break;
				case 1:	// tangents
					n = vertices.tangents[m];
					break;
				case 2:	// bitangents
					n = vertices.bitangents[m];
					break;
			}

			n = VuoPoint4d_subtract(v, VuoPoint4d_multiply(n, length));

			glVertex3d(v.x, v.y, v.z);
			glVertex3d(n.x, n.y, n.z);
		}
	}
	glEnd();

	glLoadIdentity();

	dispatch_semaphore_signal(sceneRenderer->scenegraphSemaphore);
}

/**
 * Changes the scenegraph to be rendered.
 *
 * May be called from any thread (doesn't require a GL Context).
 */
void VuoSceneRenderer_setRootSceneObject(VuoSceneRenderer sr, VuoSceneObject rootSceneObject)
{
	struct VuoSceneRendererInternal *sceneRenderer = (struct VuoSceneRendererInternal *)sr;

	dispatch_semaphore_wait(sceneRenderer->scenegraphSemaphore, DISPATCH_TIME_FOREVER);
	sceneRenderer->scenegraphChanged = true;
	VuoSceneObject_retain(rootSceneObject);
	if (sceneRenderer->scenegraphValid)
		VuoSceneObject_release(sceneRenderer->rootSceneObject);
	sceneRenderer->rootSceneObject = rootSceneObject;
	sceneRenderer->scenegraphValid = true;
	dispatch_semaphore_signal(sceneRenderer->scenegraphSemaphore);
}

/**
 * Destroys and deallocates the scene renderer.
 *
 * May be called from any thread (doesn't require a GL Context).
 */
void VuoSceneRenderer_destroy(VuoSceneRenderer sr)
{
	struct VuoSceneRendererInternal *sceneRenderer = (struct VuoSceneRendererInternal *)sr;

	dispatch_release(sceneRenderer->scenegraphSemaphore);
	VuoSceneObject_release(sceneRenderer->rootSceneObject);

	free(sceneRenderer);
}
